function getClass(byteArray, position) {
    var cls = (byteArray[position] & 0xc0) / 64;
    // Consumes no bytes
    return cls;
}

function getStructured(byteArray) {
    var structured = ((byteArray[0] & 0x20) === 0x20);
    // Consumes no bytes
    return structured;
}

function getTag(byteArray, position) {
    var tag = byteArray[0] & 0x1f;
    position += 1;
    if (tag === 0x1f) {
        tag = 0;
        while (byteArray[position] >= 0x80) {
            tag = tag * 128 + byteArray[position] - 0x80;
            position += 1;
        }
        tag = tag * 128 + byteArray[position] - 0x80;
        position += 1;
    }
    return [tag, position];
}

function getLength(byteArray, position) {
    var length = 0;

    if (byteArray[position] < 0x80) {
        length = byteArray[position];
        position += 1;
    } else {
        var numberOfDigits = byteArray[position] & 0x7f;
        position += 1;
        length = 0;
        for (var i=0; i<numberOfDigits; i++) {
            length = length * 256 + byteArray[position];
            position += 1;
        }
    }
    return [length, position];
}

function berToJavaScript(byteArray) {
    "use strict";
    var result = {};
    var position = 0;

    result.cls = getClass(byteArray, position);
    result.structured = getStructured(byteArray);
    var res = getTag(byteArray, position);
    result.tag = res[0];
    position = res[1];
    res = getLength(byteArray, position);
    var length = res[0];
    position = res[1];

    if (length === 0x80) {
        length = 0;
        while (byteArray[position + length] !== 0 || byteArray[position + length + 1] !== 0) {
            length += 1;
        }
        result.byteLength   = position + length + 2;
        result.contents     = byteArray.subarray(position, position + length);
    } else {
        result.byteLength   = position + length;
        result.contents     = byteArray.subarray(position, result.byteLength);
    }

    result.raw              = byteArray.subarray(0, result.byteLength); // May not be the whole input array
    return result;

    // Define the "get" functions here
}

function decodeBase64(encoded) {
  var decoded = window.atob(encoded);
  var der = new Uint8Array(decoded.length);
  for (var i=0; i<decoded.length; i++) {
    der[i] = decoded.charCodeAt(i);
  }
  return der;
}

function decodeCert(pem) {
  var lines = pem.split('\n');
  var encoded = '';
  for(var i = 0;i < lines.length;i++){
    if (lines[i].trim().length > 0 &&
        lines[i].indexOf('-BEGIN CERTIFICATE-') < 0 &&
        lines[i].indexOf('-END CERTIFICATE-') < 0) {
      encoded += lines[i].trim();
    }
  }
  return decodeBase64(encoded);
}

function berListToJavaScript(byteArray) {
    var result = new Array();
    var nextPosition = 0;
    while (nextPosition < byteArray.length) {
        var nextPiece = berToJavaScript(byteArray.subarray(nextPosition));
        result.push(nextPiece);
        nextPosition += nextPiece.byteLength;
    }
    return result;
}

function berObjectIdentifierValue(byteArray) {
    var oid = Math.floor(byteArray[0] / 40) + "." + byteArray[0] % 40;
    var position = 1;
    while(position < byteArray.length) {
        var nextInteger = 0;
        while (byteArray[position] >= 0x80) {
            nextInteger = nextInteger * 0x80 + (byteArray[position] & 0x7f);
            position += 1;
        }
        nextInteger = nextInteger * 0x80 + byteArray[position];
        position += 1;
        oid += "." + nextInteger;
    }
    return oid;
}

function parseAlgorithmIdentifier(asn1) {
    if (asn1.cls !== 0 || asn1.tag !== 16 || !asn1.structured) {
        throw new Error("Bad algorithm identifier. Not a SEQUENCE.");
    }
    var alg = {asn1: asn1};
    var pieces = berListToJavaScript(asn1.contents);
    if (pieces.length > 2) {
        throw new Error("Bad algorithm identifier. Contains too many child objects.");
    }
    var encodedAlgorithm = pieces[0];
    if (encodedAlgorithm.cls !== 0 || encodedAlgorithm.tag !== 6 || encodedAlgorithm.structured) {
        throw new Error("Bad algorithm identifier. Does not begin with an OBJECT IDENTIFIER.");
    }
    alg.algorithm = berObjectIdentifierValue(encodedAlgorithm.contents);
    if (pieces.length === 2) {
        alg.parameters = {asn1: pieces[1]}; // Don't need this now, so not parsing it
    } else {
        alg.parameters = null;  // It is optional
    }
    return alg;
}

function berBitStringValue(byteArray) {
    return {
        unusedBits: byteArray[0],
        bytes: byteArray.subarray(1)
    };
}

function parseSubjectPublicKeyInfo(asn1) {
    if (asn1.cls !== 0 || asn1.tag !== 16 || !asn1.structured) {
        throw new Error("Bad SPKI. Not a SEQUENCE.");
    }
    var spki = {asn1: asn1};
    var pieces = berListToJavaScript(asn1.contents);
    if (pieces.length !== 2) {
        throw new Error("Bad SubjectPublicKeyInfo. Wrong number of child objects.");
    }
    spki.algorithm = parseAlgorithmIdentifier(pieces[0]);
    spki.bits = berBitStringValue(pieces[1].contents);
    return spki;
}

function parseTBSCertificate(asn1) {
    if (asn1.cls !== 0 || asn1.tag !== 16 || !asn1.structured) {
        throw new Error("This can't be a TBSCertificate. Wrong data type.");
    }
    var tbs = {asn1: asn1};  // Include the raw parser result for debugging
    var pieces = berListToJavaScript(asn1.contents);
    if (pieces.length < 7) {
        throw new Error("Bad TBS Certificate. There are fewer than the seven required children.");
    }
    tbs.version = pieces[0];
    tbs.serialNumber = pieces[1];
    tbs.signature = parseAlgorithmIdentifier(pieces[2]);
    tbs.issuer = pieces[3];
    tbs.validity = pieces[4];
    tbs.subject = pieces[5];
    tbs.subjectPublicKeyInfo = parseSubjectPublicKeyInfo(pieces[6]);
    return tbs;  // Ignore optional fields for now
}

function parseSignatureValue(asn1) {
    if (asn1.cls !== 0 || asn1.tag !== 3 || asn1.structured) {
        throw new Error("Bad signature value. Not a BIT STRING.");
    }
    var sig = {asn1: asn1};   // Useful for debugging
    sig.bits = berBitStringValue(asn1.contents);
    return sig;
}

function parseCertificate(byteArray) {
    var asn1 = berToJavaScript(byteArray);
    if (asn1.cls !== 0 || asn1.tag !== 16 || !asn1.structured) {
        throw new Error("This can't be an X.509 certificate. Wrong data type.");
    }

    var cert = {asn1: asn1};  // Include the raw parser result for debugging
    var pieces = berListToJavaScript(asn1.contents);
    if (pieces.length !== 3) {
        throw new Error("Certificate contains more than the three specified children.");
    }

    cert.tbsCertificate     = parseTBSCertificate(pieces[0]);
    cert.signatureAlgorithm = parseAlgorithmIdentifier(pieces[1]);
    cert.signatureValue     = parseSignatureValue(pieces[2]);

    return cert;
}


function validate(certificate, text, signature) {
  var publicKey;
  var hashName = "SHA-256";
  window.crypto.subtle.importKey(
      'spki',
      certificate.tbsCertificate.subjectPublicKeyInfo.asn1.raw,
      {name: "RSASSA-PKCS1-v1_5", hash: {name: hashName}},
      true,
      ["verify"]
  ).
  then(function(key) {
      publicKey = key;
      window.crypto.subtle.verify(
          {name: "RSASSA-PKCS1-v1_5", hash: {name: hashName}},
          publicKey,
          signature,
          text
      ).
      then(function(verified) {
          if (verified) {
              alert("The signature is valid.");
          } else {
              alert("The signature is not valid.");
          }
      }).
      catch(function(err) {
          alert("Error verifying signature: " + err.message);
      });
  }).
  catch(function(err) {
      alert("Import failed: " + err.message);
  });

}

function getBytes(string) {
  var bytes = [];
  for (var i = 0; i < string.length; ++i) {
    bytes.push(string.charCodeAt(i));
  }
  return new Uint8Array(bytes);
};

function validateSelfSignature(certificate) {
  var publicKey;

  var alg = certificate.tbsCertificate.signature.algorithm;
  if (alg !== "1.2.840.113549.1.1.5" && alg !== "1.2.840.113549.1.1.11") {
     throw new Error("Signature algorithm " + alg + " is not supported yet.");
  }

  var hashName = "SHA-1";
  if (alg === "1.2.840.113549.1.1.11") {
      hashName = "SHA-256";
  }

  window.crypto.subtle.importKey(
      'spki',
      certificate.tbsCertificate.subjectPublicKeyInfo.asn1.raw,
      {name: "RSASSA-PKCS1-v1_5", hash: {name: hashName}},
      true,
      ["verify"]
  ).
  then(function(key) {
      publicKey = key;
      window.crypto.subtle.verify(
          {name: "RSASSA-PKCS1-v1_5", hash: {name: hashName}},
          publicKey,
          certificate.signatureValue.bits.bytes,
          certificate.tbsCertificate.asn1.raw
      ).
      then(function(verified) {
          if (verified) {
              alert("The certificate is properly self-signed.");
          } else {
              alert("The self-signed certificate's signature is not valid.");
          }
      }).
      catch(function(err) {
          alert("Error verifying signature: " + err.message);
      });
  }).
  catch(function(err) {
      alert("Import failed: " + err.message);
  });

}

function validateInternal() {
  var cert = $('#certText').val().trim();
  var sign = $('#signText').val().trim();
  var text = $('#text').val().trim();
  var der = decodeCert(cert);
  var certificate = parseCertificate(der);
  var signature = decodeBase64(sign);
  validate(certificate, getBytes(text), signature);
}

function loadCertificate() {
    $.get( "ws/webcrypto/certificate", function(data) {
            $('#certText').val(data);
        }
    );
}

function signExternal() {
    if ($('#text').val()) {
        $.ajax({
            type: "POST",
            url: "ws/webcrypto/sign",
            data: { text: $('#text').val()}
        })
        .done(function (data) {
            $('#signText').val(data);
        });
    }
}

function validateExternal() {
    if ($('#text').val() && $('#signText').val()) {
        $.ajax({
            type: "POST",
            url: "ws/webcrypto/validate",
            data: { text: $('#text').val(), signature: $('#signText').val()}
        })
        .done(function (data) {
            if (data === 'true') {
                alert("The signature is valid.");
            } else {
                alert("The signature is not valid.");
            }
        });
    }
}

function checkCrypto() {
    if (!window.crypto.subtle) {
        $('#valIntBtn').prop("disabled",true);
        $('#errorSpan').html('Sorry but Web Crypto Api is not supported in this browser!');
    }
}